@@ -1,9 +1,10 @@
 /*
  * tools.c
  *
- * This file is part of msmtp, an SMTP client.
+ * This file is part of msmtp, an SMTP client, and of mpop, a POP3 client.
  *
- * Copyright (C) 2004, 2005, 2006, 2007, 2009, 2011, 2014
+ * Copyright (C) 2004, 2005, 2006, 2007, 2009, 2011, 2014, 2018, 2019, 2020,
+ * 2021, 2022, 2023
  * Martin Lambers <marlam@marlam.de>
  *
  *   This program is free software; you can redistribute it and/or modify
@@ -24,19 +25,11 @@
 # include "config.h"
 #endif
 
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
-#include <errno.h>
-#include <fcntl.h>
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <time.h>
-#include <unistd.h>
 #ifdef W32_NATIVE
 # define WIN32_LEAN_AND_MEAN    /* do not include more than necessary */
-# define _WIN32_WINNT 0x0502    /* Windows XP SP2 or later */
+# define _WIN32_WINNT 0x0601    /* Windows 7 or later */
 # include <windows.h>
+# include <winsock2.h>
 # include <io.h>
 # include <conio.h>
 # include <lmcons.h>
@@ -45,11 +38,132 @@
 #else
 # include <pwd.h>
 #endif
+#include <stdio.h>
+#include <stdlib.h>
+#include <ctype.h>
+#include <string.h>
+#include <strings.h>
+#include <errno.h>
+#include <fcntl.h>
+#include <sys/stat.h>
+#include <time.h>
+#include <unistd.h>
+#ifdef ENABLE_NLS
+# include <locale.h>
+#endif
 
 #include "xalloc.h"
 #include "tools.h"
 
 
+/*
+ * exitcode_to_string()
+ *
+ * see tools.h
+ */
+
+const char *exitcode_to_string(int exitcode)
+{
+    switch (exitcode)
+    {
+        case EX_OK:
+            return "EX_OK";
+        case EX_USAGE:
+            return "EX_USAGE";
+        case EX_DATAERR:
+            return "EX_DATAERR";
+        case EX_NOINPUT:
+            return "EX_NOINPUT";
+        case EX_NOUSER:
+            return "EX_NOUSER";
+        case EX_NOHOST:
+            return "EX_NOHOST";
+        case EX_UNAVAILABLE:
+            return "EX_UNAVAILABLE";
+        case EX_SOFTWARE:
+            return "EX_SOFTWARE";
+        case EX_OSERR:
+            return "EX_OSERR";
+        case EX_OSFILE:
+            return "EX_OSFILE";
+        case EX_CANTCREAT:
+            return "EX_CANTCREAT";
+        case EX_IOERR:
+            return "EX_IOERR";
+        case EX_TEMPFAIL:
+            return "EX_TEMPFAIL";
+        case EX_PROTOCOL:
+            return "EX_PROTOCOL";
+        case EX_NOPERM:
+            return "EX_NOPERM";
+        case EX_CONFIG:
+            return "EX_CONFIG";
+        default:
+            return "BUG:UNKNOWN";
+    }
+}
+
+
+/*
+ * tmpfile() for Windows
+ *
+ * The native tmpfile() on Windows puts files in the root directory and
+ * therefore requires privileges. This is a replacement for that nonsense.
+ */
+#ifdef W32_NATIVE
+FILE *w32_tmpfile()
+{
+    static char prefix[4] = { '\0', '\0', '\0', '\0' };
+    if (prefix[0] == '\0')
+    {
+        /* initialize the prefix once per process so that different processes use
+         * different prefixes, thus reducing the chances of file name collisions
+         * in the loop below */
+        DWORD pid = GetCurrentProcessId();
+        prefix[0] = 'a' + pid % 26;
+        prefix[1] = 'a' + (pid / 26) % 26;
+        prefix[2] = 'a' + (pid / (26 * 26)) % 26;
+    }
+    /* First get a name. Unfortunately Windows _tempnam() only looks at $TMP
+     * but not at system default destinations, thus we also have to use GetTempPathW().
+     * Furthermore, _tempnam() might return a file name prepended with a backslash to
+     * mean the current directory, so we have to clean that up. */
+    char dirname[MAX_PATH + 2];
+    int fd = -1;
+    int i = 0;
+    do
+    {
+        i++;
+        if (i > TMP_MAX)
+        {
+            errno = EEXIST;
+            return NULL;
+        }
+        DWORD r = GetTempPath(sizeof(dirname), dirname);
+        char *buf = _tempnam(r == 0 ? NULL : dirname, prefix);
+        char *name = buf;
+        if (!name)
+        {
+            errno = EEXIST;
+            return NULL;
+        }
+        if (name[0] == '\\' && !strchr(name + 1, '\\'))
+        {
+            name++;
+        }
+        /* Now create the file with O_EXCL to avoid race conditions. */
+        fd = _open(name, _O_RDWR
+                | _O_CREAT | _O_TRUNC | _O_EXCL
+                | _O_TEMPORARY
+                | _O_BINARY,
+                _S_IREAD | _S_IWRITE);
+        free(buf);
+    }
+    while (fd < 0);
+    return (fd >= 0 ? fdopen(fd, "w+b") : NULL);
+}
+#endif
+
 /*
  * link()
  *
@@ -229,7 +343,7 @@ char *get_sysconfdir(void)
     RegCloseKey(hkey);
     return xstrdup((char *)sysconfdir);
 
-#else /* UNIX or DJGPP */
+#else /* UNIX */
 
 #ifdef SYSCONFDIR
     return xstrdup(SYSCONFDIR);
@@ -253,7 +367,6 @@ char *get_username(void)
 #ifdef W32_NATIVE
     DWORD size = UNLEN + 1;
     TCHAR buf[UNLEN + 1];
-#elif defined DJGPP
 #else /* UNIX */
     struct passwd *pw;
 #endif
@@ -282,12 +395,6 @@ char *get_username(void)
                 /* last resort */
                 username = xstrdup("unknown");
             }
-#elif defined DJGPP
-            /* DJGPP's getlogin() checks USER, then LOGNAME, and then uses
-             * "dosuser" as a last resort. We already checked USER and LOGNAME
-             * and choose "unknown" as a last resort to be consistent with the
-             * other systems. */
-            username = xstrdup("unknown");
 #else /* UNIX */
             username = getlogin();
             if (username)
@@ -315,6 +422,83 @@ char *get_username(void)
 }
 
 
+/*
+ * get_hostname()
+ *
+ * see tools.h
+ */
+
+char *get_hostname(void)
+{
+    char *host;
+
+    host = getenv("HOSTNAME");
+    if (host)
+    {
+        host = xstrdup(host);
+    }
+    else
+    {
+        char buf[256];
+        if (gethostname(buf, 256) == 0)
+        {
+            /* Make sure the hostname is NUL-terminated. */
+            buf[255] = '\0';
+            host = xstrdup(buf);
+        }
+    }
+    if (!host)
+    {
+        host = xstrdup("localhost");
+    }
+    return host;
+}
+
+
+/*
+ * get_userconfig()
+ *
+ * see tools.h
+ */
+
+char *get_userconfig(const char *userconfigfile)
+{
+    char *homedir = get_homedir();
+    char *path = get_filename(homedir, userconfigfile);
+
+#if !defined(W32_NATIVE)
+    struct stat buf;
+    char *xdg_home;
+    char *newpath;
+
+    //does not exist, thus check XDG_CONFIG_HOME/PACKAGE_NAME/config
+    if (stat(path, &buf) != 0) {
+        xdg_home = getenv("XDG_CONFIG_HOME");
+        if (xdg_home) {
+            xdg_home = xstrdup(xdg_home);
+        } else {
+            xdg_home = expand_tilde("~/.config");
+        }
+        newpath = get_filename(xdg_home, PACKAGE_NAME);
+        free(xdg_home);
+        xdg_home = get_filename(newpath, "config");
+        free(newpath);
+        newpath = xdg_home;
+        //If this does not exist fallback
+        if (stat(newpath, &buf) == 0) {
+            free(path);
+            path = newpath;
+        } else {
+            free(newpath);
+        }
+    }
+#endif
+
+    free(homedir);
+    return path;
+}
+
+
 /*
  * get_homedir()
  *
@@ -360,21 +544,6 @@ char *get_homedir(void)
 
     return home;
 
-#elif defined DJGPP
-
-    char *home;
-
-    if ((home = getenv("HOME")))
-    {
-        home = xstrdup(home);
-    }
-    else
-    {
-        home = xstrdup("C:");
-    }
-
-    return home;
-
 #else /* UNIX */
 
     char *home;
@@ -462,7 +631,7 @@ char *expand_tilde(const char *filename)
 
 int check_secure(const char *pathname)
 {
-#if defined W32_NATIVE || defined DJGPP || defined __CYGWIN__
+#if defined W32_NATIVE || defined __CYGWIN__
 
     return 0;
 
@@ -491,202 +660,6 @@ int check_secure(const char *pathname)
 }
 
 
-/*
- * [DJGPP and Windows only] mymkstemp()
- *
- * This function does on DOS/Windows what mkstemp() followed by unlink() do on
- * UNIX.
- *
- * 1. unlink() on DOS and Windows is not POSIX conformant: it does not wait
- *    until the last file descriptor is closed before unlinking the file.
- *    Instead, it fails (Windows) or may even mess up the file system (DOS).
- * 2. Windows does not have mkstemp.
- * 3. If a file is opened with O_TEMPORARY on Windows or DOS, it will be deleted
- *    after the last file descriptor is closed. This is what this function does.
- *
- * Return value: file descriptor, or -1 on error (errno will be set).
- */
-
-#if defined(W32_NATIVE) || defined(DJGPP)
-int mymkstemp(char *template, int remove_on_close)
-{
-    size_t templatelen;
-    char *X;
-    int i;
-    int try;
-    int ret;
-    const char alnum[]
-        = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
-#ifdef W32_NATIVE
-    int temporary_flag = (remove_on_close ? _O_TEMPORARY : 0);
-#else /* DJGPP */
-    int temporary_flag = (remove_on_close ? O_TEMPORARY : 0);
-#endif /* DJGPP */
-
-    templatelen = strlen(template);
-    if (templatelen < 6)
-    {
-        errno = EINVAL;
-        return -1;
-    }
-    X = template + templatelen - 6;
-    if (strcmp(X, "XXXXXX") != 0)
-    {
-        errno = EINVAL;
-        return -1;
-    }
-
-    srand((unsigned int)time(NULL));
-
-    /* We have 62^6 possible filenames. We try 62^2=3844 times. */
-    ret = -1;
-    for (try = 0; ret == -1 && try < 3844; try++)
-    {
-        for (i = 0; i < 6; i++)
-        {
-            X[i] = alnum[rand() % 36];
-        }
-#ifdef W32_NATIVE
-        ret = _open(template, _O_CREAT | _O_EXCL | _O_RDWR | _O_BINARY
-                | temporary_flag, _S_IREAD | _S_IWRITE);
-#else /* DJGPP */
-        ret = open(template, O_CREAT | O_EXCL | O_RDWR | _O_BINARY
-                | temporary_flag, S_IRUSR | S_IWUSR);
-#endif /* DJGPP */
-    }
-
-    return ret;
-}
-#endif /* W32_NATIVE or DJGPP */
-
-#ifndef HAVE_MKSTEMP
-# if defined(W32_NATIVE) || defined(DJGPP)
-int mkstemp(char *template)
-{
-    return mymkstemp(template, 0);
-}
-# endif
-#endif
-
-
-/*
- * tempfile()
- *
- * see tools.h
- */
-
-FILE *tempfile(const char *base)
-{
-    FILE *f;
-    size_t baselen;
-    const char *dir;
-    size_t dirlen;
-    char *template = NULL;
-    size_t templatelen;
-    int fd = -1;
-    int saved_errno;
-
-    if (!base || (*base == '\0'))
-    {
-        base = "tmp";
-    }
-    /* the directory for the temp file */
-    if (!(dir = getenv("TMPDIR")))
-    {
-        /* system dependent default location */
-#ifdef W32_NATIVE
-        /* there is no registry key for this (?) */
-        if (!(dir = getenv("TEMP")))
-        {
-            if (!(dir = getenv("TMP")))
-            {
-                dir = "C:";
-            }
-        }
-#elif defined DJGPP
-        dir = "C:";
-#else /* UNIX */
-#ifdef P_tmpdir
-        dir = P_tmpdir;
-#else
-        dir = "/tmp";
-#endif
-#endif /* UNIX */
-    }
-    dirlen = strlen(dir);
-
-    /* the proposed file name */
-    baselen = strlen(base);
-#ifdef DJGPP
-    /* shorten the base to two characters because of 8.3 filenames */
-    if (baselen > 2)
-    {
-        baselen = 2;
-    }
-#endif
-
-    /* build the template */
-    templatelen = dirlen + 1 + baselen + 6;
-    template = xmalloc((templatelen + 1) * sizeof(char));
-    strncpy(template, dir, dirlen);
-    if (dirlen == 0 || template[dirlen - 1] != PATH_SEP)
-    {
-        template[dirlen++] = PATH_SEP;
-    }
-    /* template is long enough */
-    strncpy(template + dirlen, base, baselen);
-    strcpy(template + dirlen + baselen, "XXXXXX");
-
-    /* create the file */
-#if defined(W32_NATIVE) || defined(DJGPP)
-    if ((fd = mymkstemp(template, 1)) == -1)
-#else /* UNIX */
-    if ((fd = mkstemp(template)) == -1)
-#endif /* UNIX */
-    {
-        goto error_exit;
-    }
-
-    /* UNIX only: set the permissions (not every mkstemp() sets them to 0600)
-     * and unlink the file so that it gets deleted when the caller closes it */
-#ifndef DJGPP
-#ifndef W32_NATIVE
-    if (fchmod(fd, S_IRUSR | S_IWUSR) == -1)
-    {
-        goto error_exit;
-    }
-    if (unlink(template) != 0)
-    {
-        goto error_exit;
-    }
-#endif /* not W32_NATIVE */
-#endif /* not DJGPP */
-
-    /* get the stream from the filedescriptor */
-    if (!(f = fdopen(fd, "w+")))
-    {
-        goto error_exit;
-    }
-    free(template);
-
-    return f;
-
-error_exit:
-    saved_errno = errno;
-    if (fd >= 0)
-    {
-        close(fd);
-    }
-    if (template)
-    {
-        (void)remove(template);
-        free(template);
-    }
-    errno = saved_errno;
-    return NULL;
-}
-
-
 /*
  * lock_file()
  *
@@ -698,8 +671,6 @@ static void sleep_tenth_second(void)
 {
 #ifdef W32_NATIVE
     Sleep(100);
-#elif defined DJGPP
-    usleep(100000);
 #else /* POSIX */
     struct timespec tenth_second = { 0, 100000000 };
     nanosleep(&tenth_second, NULL);
@@ -728,7 +699,7 @@ int lock_file(FILE *f, int lock_type, int timeout)
         errno = 0;
 #ifdef W32_NATIVE
         lock_success = (_locking(fd, _LK_NBLCK, LONG_MAX) != -1);
-#else /* UNIX, DJGPP */
+#else /* UNIX */
         lock_success = (fcntl(fd, F_SETLK, &lock) != -1);
 #endif
         if (lock_success || (errno != EACCES && errno != EAGAIN)
@@ -773,6 +744,81 @@ char *string_replace(char *str, const char *s, const char *r)
     return str;
 }
 
+
+/*
+ * sanitize_string()
+ *
+ * see tools.h
+ */
+
+char *sanitize_string(char *str)
+{
+    char *p = str;
+
+    while (*p != '\0')
+    {
+        if (iscntrl((unsigned char)*p))
+        {
+            *p = '?';
+        }
+        p++;
+    }
+
+    return str;
+}
+
+/*
+ * token_in_string()
+ *
+ * see tools.h
+ */
+
+int token_in_string(const char *string, const char *token)
+{
+    size_t token_len = strlen(token);
+    const char *oldstring = string;
+    const char *newstring;
+    int found = 0;
+
+    while ((newstring = strstr(oldstring, token)))
+    {
+        size_t i = newstring - string;
+        if (i == 0 || string[i - 1] == ' ') /* valid start of token */
+        {
+            if (string[i + token_len] == ' '
+                    || string[i + token_len] == '\r'
+                    || string[i + token_len] == '\0') /* valid end of token */
+            {
+                found = 1;
+                break;
+            }
+        }
+        oldstring = newstring + 1;
+    }
+    return found;
+}
+
+
+/*
+ * print_fingerprint()
+ *
+ * see tools.h
+ */
+
+void print_fingerprint(char *s, const unsigned char *fingerprint, size_t len)
+{
+    const char *hex = "0123456789ABCDEF";
+    size_t i;
+
+    for (i = 0; i < len; i++)
+    {
+        s[3 * i + 0] = hex[(fingerprint[i] & 0xf0) >> 4];
+        s[3 * i + 1] = hex[fingerprint[i] & 0x0f];
+        s[3 * i + 2] = (i < len - 1 ? ':' : '\0');
+    }
+}
+
+
 /*
  * print_time_rfc2822()
  *
@@ -781,47 +827,83 @@ char *string_replace(char *str, const char *s, const char *r)
 
 void print_time_rfc2822(time_t t, char rfc2822_timestamp[32])
 {
-    /* Calculate a RFC 2822 timestamp. strftime() is unreliable for this because
-     * it is locale dependant, and because the timezone offset conversion
-     * specifier %z is not portable. */
-    struct tm gmt, *lt;
-    char tz_offset_sign;
-    int tz_offset_hours;
-    int tz_offset_minutes;
-    const char *weekday[7] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri",
-        "Sat" };
-    const char *month[12] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
-        "Aug", "Sep", "Oct", "Nov", "Dec" };
+    struct tm *lt = localtime(&t);
+#ifdef ENABLE_NLS
+    /* Set the correct locale for strftime() */
+    char *old_locale, *saved_locale;
+    old_locale = setlocale(LC_ALL, NULL);
+    saved_locale = xstrdup(old_locale);
+    setlocale(LC_ALL, "C");
+#endif
+    strftime(rfc2822_timestamp, 32, "%a, %d %b %Y %T %z", lt);
+#ifdef ENABLE_NLS
+    /* Restore the original locale */
+    setlocale(LC_ALL, saved_locale);
+    free(saved_locale);
+#endif
+}
 
-    /* copy the struct tm, because the subsequent call to localtime() will
-     * overwrite it */
-    gmt = *gmtime(&t);
-    lt = localtime(&t);
-    tz_offset_minutes = (lt->tm_hour - gmt.tm_hour) * 60
-        + lt->tm_min - gmt.tm_min
-        + (lt->tm_year - gmt.tm_year) * 24 * 60
-        + (lt->tm_yday - gmt.tm_yday) * 24 * 60;
-    if (tz_offset_minutes < 0)
+
+/*
+ * split_mail_address()
+ *
+ * see tools.h
+ */
+
+void split_mail_address(const char *address, char **local_part, char **domain_part)
+{
+    const char *p = strrchr(address, '@');
+    if (p)
     {
-        tz_offset_sign = '-';
-        tz_offset_minutes = -tz_offset_minutes;
+        size_t local_part_len = p - address;
+        size_t domain_part_len = strlen(p + 1);
+        *local_part = xmalloc(local_part_len + 1);
+        strncpy(*local_part, address, local_part_len);
+        (*local_part)[local_part_len] = '\0';
+        *domain_part = xmalloc(domain_part_len + 1);
+        strcpy(*domain_part, p + 1);
     }
     else
     {
-        tz_offset_sign = '+';
+        size_t local_part_len = strlen(address);
+        *local_part = xmalloc(local_part_len + 1);
+        strcpy(*local_part, address);
+        *domain_part = NULL;
     }
-    tz_offset_hours = tz_offset_minutes / 60;
-    tz_offset_minutes %= 60;
-    if (tz_offset_hours > 99)
-    {
-        /* Values equal to or larger than 24 are not meaningful, but we just
-         * make sure that the value fits into two digits. If the system time is
-         * broken, we cannot fix it. */
-        tz_offset_hours = 99;
-    }
-    (void)snprintf(rfc2822_timestamp, 32,
-            "%s, %02d %s %04d %02d:%02d:%02d %c%02d%02d",
-            weekday[lt->tm_wday], lt->tm_mday, month[lt->tm_mon],
-            lt->tm_year + 1900, lt->tm_hour, lt->tm_min, lt->tm_sec,
-            tz_offset_sign, tz_offset_hours, tz_offset_minutes);
+}
+
+
+/*
+ * check_hostname_matches_domain()
+ *
+ * see tools.h
+ */
+
+int check_hostname_matches_domain(const char *hostname, const char *domain)
+{
+    size_t hostname_len = strlen(hostname);
+    size_t domain_len = strlen(domain);
+
+    /* empty domain? */
+    if (domain_len < 1)
+        return 0;
+
+    /* host name shorter than domain? */
+    if (hostname_len < domain_len)
+        return 0;
+
+    /* if lengths match, then the strings must match */
+    if (hostname_len == domain_len)
+        return strcasecmp(hostname, domain) == 0 ? 1 : 0;
+
+    /* if host name is longer, than it must be at least two longer because of
+     * the '.' (e.g. hostname="a.example.com", domain="example.com") */
+    if (hostname_len < domain_len + 2)
+        return 0;
+
+    /* host name is at least two longer than domain name:
+     * check that domain matches and that we have a separating dot
+     * (so that hostname="xxexample.com" does not match "example.com") */
+    return (hostname[hostname_len - 1 - domain_len] == '.'
+            && strcasecmp(hostname + (hostname_len - domain_len), domain) == 0) ? 1 : 0;
 }
